package be.iminds.ilabt.jfed.espec.util;

import be.iminds.ilabt.jfed.espec.model.*;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.util.List;

/**
 * ESpecLogListener is used to get feedback about all ExperimentSpecification actions and steps
 *
 * Do not expect to be all calls to be on the same thread!
 * This means that you might theoretically get some of the onPostExecute() calls after the onPostExecuteAll() call...
 */
public interface ESpecLogListener {
    class ESpecStepStatus {
        private final boolean success;
        @Nullable private final String message;
        @Nullable private final Throwable throwable;

        public ESpecStepStatus(boolean success) {
            this.success = success;
            this.message = null;
            this.throwable = null;
        }
        public ESpecStepStatus(boolean success, @Nullable String message, @Nullable Throwable throwable) {
            this.success = success;
            this.message = message;
            this.throwable = throwable;
        }

        public boolean isSuccess() {
            return success;
        }
        public boolean isFailure() {
            return !success;
        }

        @Nullable
        public String getMessage() {
            return message;
        }

        @Nullable
        public Throwable getThrowable() {
            return throwable;
        }
    }

    void onPreFileLoad();
    void onPostFileLoad(@Nonnull FileSource fileSource, boolean isFast, long byteSize, @Nonnull ESpecStepStatus status);
    void onPostFileLoadAll(boolean success);

    void onPreRSpec();
    /** Called when the request RSpec to be used is known. */
    void onRequestRSpecKnown(@Nonnull RspecSpec rspecSpec, @Nonnull String requestRspec);
    void onPostRSpec(@Nonnull RspecSpec rspec, @Nonnull ESpecStepStatus status);
    void onPostRspecAll(boolean success, @Nonnull List<String> allNodeClientIds);

    void onPreDir();
    void onPreDirNode(@Nonnull String nodeClientId);
    void onPostDir(@Nonnull DirSpec dir, @Nonnull String fullPath, @Nonnull String nodeClientId,
                   @Nonnull ESpecStepStatus status);
    void onPostDirNode(@Nonnull String nodeClientId, boolean success);
    void onPostDirAll(boolean success);

    void onPreUpload();
    void onPreUploadNode(@Nonnull String nodeClientId);
    void onPostUpload(@Nonnull UploadLikeSpec uploadSpec, @Nonnull String fullPath, @Nonnull String nodeClientId,
                      @Nonnull ESpecStepStatus status);
    void onPostUploadNode(@Nonnull String nodeClientId, boolean success);
    void onPostUploadAll(boolean success);

    void onPreExecute();
    //not needed:     void onPreExecuteNode(@Nonnull String nodeClientId);
    //because executes are synchonized between nodes! (= barrier between each execute)

    /**
     * @param executeSpec the configuration of this "execute script" run
     * @param fullExecutablePath the full path of the executable script on the remote (the file was typically copied here in the upload step).
     * @param fullLogPath the full path of the log (stdout and stderr redirected to a file) of the execute on the remote. (some options can disable this logging, but it is enabled by default) Can be null if there is no lo, or if there was an error before the log was created.
     * @param nodeClientId the client_id of the node this execution ran on
     * @param exitStatus the exit status of this execute. If the executable did not run due to errors, -1 is returned.
     * @param status containing: success true if no errors occured and the exit status is 0. false otherwise.
     */
    void onPostExecute(@Nonnull ExecuteSpec executeSpec, @Nonnull String fullExecutablePath,
                       @Nullable String fullLogPath, @Nonnull String nodeClientId, int exitStatus,
                       @Nonnull ESpecStepStatus status);
    void onPostExecuteStepAll(@Nonnull ExecuteSpec executeSpec, boolean success);
    //not needed:      void onPostExecuteNode(@Nonnull String nodeClientId, boolean success);
    void onPostExecuteAll(boolean success);

    void onPreAnsible();
    /**
     * @param ansiblePlaybookSpec the configuration of this ansible run
     * @param fullPlaybookPath the full path of the ansible playbook on the remote (playbook was typically copied here in the upload step).
     * @param fullLogPath the full path of the log (stdout and stderr redirected to a file) of the playbook on the remote. (some options can disable this logging, but it is enabled by default) Can be null if there is no lo, or if there was an error before the log was created.
     * @param nodeClientId the client_id of the node ansible ran on (= ansible controller client_id).
     * @param status containing: success true if the ansible-playbook executable returned success. (You can configure custom failure conditions in the ansible playbook.)
     */
    void onPostAnsiblePlaybook(@Nonnull AnsiblePlaybookSpec ansiblePlaybookSpec, @Nonnull String fullPlaybookPath,
                               @Nullable String fullLogPath, @Nonnull String nodeClientId,
                               @Nonnull ESpecStepStatus status);
    void onPostAnsibleAll(boolean success);
}
