package be.iminds.ilabt.jfed.experimenter_gui.slice.espec.model;

import be.iminds.ilabt.jfed.espec.model.*;
import be.iminds.ilabt.jfed.espec.util.ESpecLogListener;
import be.iminds.ilabt.jfed.experiment.Experiment;
import be.iminds.ilabt.jfed.experiment.ExperimentState;
import be.iminds.ilabt.jfed.experimenter_gui.ui.status.TaskStatusIndicator;
import javafx.application.Platform;
import javafx.beans.property.ObjectProperty;
import javafx.beans.property.SimpleObjectProperty;
import javafx.collections.FXCollections;
import javafx.collections.ObservableList;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Optional;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicBoolean;

/**
 * ESpecLogModel
 */
public class ESpecLogModel implements ESpecLogListener {
    private static final org.slf4j.Logger LOG = LoggerFactory.getLogger(ESpecLogModel.class);

    public final static String AQUIRE_RESOURCES_STEP_ID = "Acquire Resources";
    public final static String SETUP_SOFTWARE_STEP_ID = ExperimentState.SETUP_SOFTWARE.name();

    private final static class ESpecStateId {
        @Nullable private final String baseId;
        @Nullable private final String subId;

        public ESpecStateId(ExperimentState experimentState, String especStepId) {
            this.baseId = findBaseId(experimentState, especStepId);
            this.subId = findSubId(experimentState, especStepId);
            LOG.debug("Mapped IDs: "+experimentState+" "+especStepId+"  -->  "+baseId+" "+subId);
        }
        private ESpecStateId(String baseId, String subId) {
            this.baseId = baseId;
            this.subId = subId;
        }

        public ESpecStateId(ExperimentState experimentState) {
            this(experimentState, null);
        }

        private static boolean partOfResourceSetup(@Nonnull ExperimentState experimentState) {
            return experimentState == ExperimentState.ALLOCATING ||
                    experimentState == ExperimentState.PROVISIONING ||
                    experimentState == ExperimentState.WAIT_FOR_READY ||
                    experimentState == ExperimentState.TESTING;
        }

        private String findBaseId(ExperimentState experimentState, String especStepId) {
            if (experimentState == null) {
                if (especStepId != null)
                    return ExperimentState.SETUP_SOFTWARE.name();
                throw new RuntimeException("Unknown state: experimentState=" + experimentState + " especStepId=" + especStepId);
            }
            if (partOfResourceSetup(experimentState))
                return AQUIRE_RESOURCES_STEP_ID;
            return experimentState.name();
        }
        public String findSubId(ExperimentState experimentState, String especStepId) {
            if (experimentState == null) {
                if (especStepId != null)
                    return especStepId;
                throw new RuntimeException("Unknown state: experimentState=" + experimentState + " especStepId=" + especStepId);
            }
            if (partOfResourceSetup(experimentState))
                return experimentState.name();

            return especStepId;
        }

        public String getBaseId() {
            return baseId;
        }
        public String getSubId() {
            return subId;
        }
        public boolean isChild() {
            return subId != null;
        }
        public ESpecStateId getParentId() {
            return new ESpecStateId(baseId, null);
        }

        @Override
        public boolean equals(Object o) {
            if (this == o) return true;
            if (o == null || getClass() != o.getClass()) return false;
            ESpecStateId that = (ESpecStateId) o;
            return Objects.equals(baseId, that.baseId) &&
                    Objects.equals(subId, that.subId);
        }

        @Override
        public int hashCode() {
            return Objects.hash(baseId, subId);
        }

        @Override
        public String toString() {
            return "ESpecStateId{" + baseId + ", " + subId + '}';
        }
    }

    private final Map<ESpecStateId, ESpecStateItem> internalESpecStateItems = new ConcurrentHashMap<>();

    private final ObservableList<ESpecLogItem> eSpecLogItems;
    private final ObservableList<ESpecStateItem> eSpecStateItems;

    private final ObjectProperty<ESpecStateItem> selectedESpecStateItem;

    @Nonnull private final Experiment experiment;

    public ESpecLogModel(@Nonnull Experiment experiment) {
        this.experiment = experiment;

        this.eSpecLogItems = FXCollections.observableArrayList();
        this.eSpecStateItems = FXCollections.observableArrayList();
        this.selectedESpecStateItem = new SimpleObjectProperty<>();

        experiment.addESpecLogListener(this);
        experiment.experimentStateProperty().addListener(
                (observableValue, oldVal, newVal) -> {
                    setPreviousRunningToSuccess(newVal, null);
                    boolean success = !isFailingState(newVal);
                    ESpecStateItem.ESpecStateStatus targetState = success ? (
                            newVal.isActiveState() && newVal != ExperimentState.READY ?
                                    ESpecStateItem.ESpecStateStatus.RUNNING :
                                    ESpecStateItem.ESpecStateStatus.SUCCESS) :
                            ESpecStateItem.ESpecStateStatus.FAILURE;
                    ESpecStateItem cur = getState(newVal, null, targetState);
                    cur.setStatus(targetState);
                    addLog()
                            .setText("Reached State " + newVal.name())
                            .setStatus(success)
                            .setESpecStateItem(cur)
                            .add();
                }
        );

        addInitialStates();
    }

    private void addInitialStates() {
        getState(ExperimentState.ALLOCATING, null, ESpecStateItem.ESpecStateStatus.FUTURE);
        getState(ExperimentState.PROVISIONING, null, ESpecStateItem.ESpecStateStatus.FUTURE);
        getState(ExperimentState.WAIT_FOR_READY, null, ESpecStateItem.ESpecStateStatus.FUTURE);
        getState(ExperimentState.TESTING, null, ESpecStateItem.ESpecStateStatus.FUTURE);
        getState(ExperimentState.SETUP_SOFTWARE, null, ESpecStateItem.ESpecStateStatus.FUTURE);
        getState(ExperimentState.SETUP_SOFTWARE, "Dir", ESpecStateItem.ESpecStateStatus.FUTURE);
        getState(ExperimentState.SETUP_SOFTWARE, "Upload", ESpecStateItem.ESpecStateStatus.FUTURE);
        getState(ExperimentState.SETUP_SOFTWARE, "Execute", ESpecStateItem.ESpecStateStatus.FUTURE);
        getState(ExperimentState.SETUP_SOFTWARE, "Ansible", ESpecStateItem.ESpecStateStatus.FUTURE);
        getState(ExperimentState.READY, null, ESpecStateItem.ESpecStateStatus.FUTURE);
    }

    /**
     * Creates the state if not existing. Thread safe.
     * JavaFX thread will automatically be notified later when needed.
     *
     * ESpecStateItem of parent and child will be created when needed.
     * */
    @Nonnull
    private ESpecStateItem getState(ExperimentState experimentState, String especStepId, ESpecStateItem.ESpecStateStatus initialStatus) {
        ESpecStateId id = new ESpecStateId(experimentState, especStepId);
        return getState(id, initialStatus);
    }
    @Nonnull
    private ESpecStateItem getState(ESpecStateId id, ESpecStateItem.ESpecStateStatus initialStatus) {
        if (id.isChild()) {
            ESpecStateId parentId = id.getParentId();
            ESpecStateItem parent = getState(parentId, initialStatus);

            AtomicBoolean wasAdded = new AtomicBoolean(false);
            ESpecStateItem res = internalESpecStateItems.computeIfAbsent(id, (key) -> {
                ESpecStateItem toAdd = new ESpecStateItem(id.subId, initialStatus);
                wasAdded.set(true);
                return toAdd;
            });
            if (wasAdded.get()) {
                LOG.debug("Added child "+res.getId()+" ("+id+")"+" to "+parent.getId()+" ("+parentId+")");
                Platform.runLater(() -> {
                    parent.getChildren().add(res);
                });
            }
            return res;
        } else {
            AtomicBoolean wasAdded = new AtomicBoolean(false);
            ESpecStateItem res = internalESpecStateItems.computeIfAbsent(id, (key) -> {
                ESpecStateItem toAdd = new ESpecStateItem(id.baseId, initialStatus);
                wasAdded.set(true);
                return toAdd;
            });
            if (wasAdded.get()) {
                LOG.debug("Added parent "+res.getId()+" ("+id+")");
                Platform.runLater(() -> {
                    eSpecStateItems.add(res);
                });
            }
            return res;
        }
    }
    private void setPreviousRunningToSuccess(ExperimentState experimentState, String especStepId) {
        setPreviousRunningToSuccess(new ESpecStateId(experimentState, especStepId));
    }
    private void setPreviousRunningToSuccess(ESpecStateId id) {
        internalESpecStateItems.values()
                .stream()
                .filter(esi -> esi.getId() != id.baseId)
                .filter(esi -> id.subId == null || esi.getId() != id.subId)
                .filter(esi -> esi.getStatus() == ESpecStateItem.ESpecStateStatus.RUNNING)
                .forEach(esi -> esi.setStatus(ESpecStateItem.ESpecStateStatus.SUCCESS));
    }

    private void setStatus(ExperimentState experimentState, String especStepId, ESpecStateItem.ESpecStateStatus newStatus) {
        ESpecStateId id = new ESpecStateId(experimentState, especStepId);
        setStatus(id, newStatus);
    }
    private void setStatus(ESpecStateId id, ESpecStateItem.ESpecStateStatus newStatus) {
        ESpecStateItem item = getState(id, newStatus);
        item.setStatus(newStatus);
    }


    public ObservableList<ESpecLogItem> getESpecLogItems() {
        return eSpecLogItems;
    }

    public ObservableList<ESpecStateItem> getESpecStateItems() {
        return eSpecStateItems;
    }

    @Nullable
    public ESpecStateItem getSelectedESpecStateItem() {
        return selectedESpecStateItem.get();
    }

    public ObjectProperty<ESpecStateItem> selectedESpecStateItemProperty() {
        return selectedESpecStateItem;
    }

    public void setSelectedESpecStateItem(@Nullable ESpecStateItem selectedESpecStateItem) {
        this.selectedESpecStateItem.set(selectedESpecStateItem);
    }
    public void clearSelectedESpecStateItem() {
        this.selectedESpecStateItem.set(null);
    }

    public static boolean isFailingState(@Nonnull ExperimentState state) {
        return state == ExperimentState.FAILING || state == ExperimentState.FAILED;
    }
    public static boolean isFailingState(@Nonnull String stateName) {
        return stateName.equalsIgnoreCase(ExperimentState.FAILING.name()) ||
                stateName.equalsIgnoreCase(ExperimentState.FAILED.name());
    }



    private class ESpecLogItemAdder {
        private String text;
        private TaskStatusIndicator.Status status;

        //info on ESpecStateItem to log to
        private ESpecStateItem eSpecStateItem;
        private ESpecStateId eSpecStateId;

        private boolean minor = false;

        ESpecLogItemAdder setText(@Nonnull String text) {
            this.text = text;
            return this;
        }

        ESpecLogItemAdder setStatus(@Nonnull TaskStatusIndicator.Status status) {
            this.status = status;
            return this;
        }
        ESpecLogItemAdder setStatus(boolean success) {
            if (success) {
                setStatus(TaskStatusIndicator.Status.SUCCESS);
            } else {
                setStatus(TaskStatusIndicator.Status.FAILED);
            }
            return this;
        }
        ESpecLogItemAdder setStatus(@Nonnull ESpecStepStatus status) {
            setStatus(status.isSuccess());
            return this;
        }

        public ESpecLogItemAdder setESpecStateId(ExperimentState experimentState, String especStepId) {
            this.eSpecStateId = new ESpecStateId(experimentState, especStepId);
            return this;
        }
        public ESpecLogItemAdder setExperimentState(ExperimentState experimentState) {
            this.eSpecStateId = new ESpecStateId(experimentState, null);
            return this;
        }
        public ESpecLogItemAdder setESpecStepId(String especStepId) {
            this.eSpecStateId = new ESpecStateId(ExperimentState.SETUP_SOFTWARE, especStepId);;
            return this;
        }
        public ESpecLogItemAdder setESpecStateId(ESpecStateId eSpecStateId) {
            this.eSpecStateId = eSpecStateId;
            return this;
        }

        ESpecLogItemAdder setESpecStateItem(ESpecStateItem eSpecStateItem) {
            this.eSpecStateItem = eSpecStateItem;
            return this;
        }

        //set base state to whatever the current base state is
        ESpecLogItemAdder setCurrentBaseState() {
            this.eSpecStateId = new ESpecStateId(experiment.getExperimentState().name(), null);
            return this;
        }

        //set base state to whatever the current base state is
        ESpecLogItemAdder setIsMinor() {
            this.minor = true;
            return this;
        }
        ESpecLogItemAdder setIsMinor(boolean newMinor) {
            this.minor = newMinor;
            return this;
        }

        public void add() {
            if (text == null) throw new IllegalStateException("text should have been set");
            if (status == null) throw new IllegalStateException("status should have been set");
            if (eSpecStateId == null && eSpecStateItem == null) throw new IllegalStateException("eSpecStateId or eSpecStateItem should have been set");
            ESpecLogItem res = new ESpecLogItem(text, status);

            if (eSpecStateId != null) {
                eSpecStateItem = getState(eSpecStateId, ESpecStateItem.ESpecStateStatus.RUNNING);
                if (eSpecStateItem == null) {
                    throw new IllegalStateException("Should never happen");
                }
            }
            if (eSpecStateItem != null) {
                res.setESpecStateItem(eSpecStateItem);
            }

            Platform.runLater(() -> {
                addESpecLogItem(res);
            });
        }
    }

    private void addESpecLogItem(@Nonnull ESpecLogItem eSpecLogItem) {
        eSpecLogItems.add(eSpecLogItem);
        if (eSpecLogItem.getESpecStateItem() != null) {
            eSpecLogItem.getESpecStateItem().getLogItems().add(eSpecLogItem);
        }
    }
    private ESpecLogItemAdder addLog() {
        return new ESpecLogItemAdder();
    }



    @Override
    public void onPreFileLoad() {

    }

    @Override
    public void onPostFileLoad(@Nonnull FileSource fileSource,
                               boolean isFast, long byteSize,
                               @Nonnull ESpecStepStatus status) {
        addLog().setText("File load \""+fileSource+"\"").setStatus(status)
                .setCurrentBaseState()
                .setIsMinor(!status.isFailure())
                .add();
    }

    @Override
    public void onPostFileLoadAll(boolean success) {
        addLog().setText("All files loaded").setStatus(success)
                .setCurrentBaseState()
                .add();
    }

    @Override
    public void onPreRSpec() {

    }

    @Override
    public void onRequestRSpecKnown(@Nonnull RspecSpec rspecSpec, @Nonnull String requestRspec) {
        addLog().setText("Request RSpec Available").setStatus(true)
                .setCurrentBaseState()
                .add();
    }

    @Override
    public void onPostRSpec(@Nonnull RspecSpec rspec, @Nonnull ESpecStepStatus status) {
        if (status.isFailure())
            addLog().setText("RSpec Provision").setStatus(status)
                    .setCurrentBaseState()
                    .add();
    }

    @Override
    public void onPostRspecAll(boolean success, @Nonnull List<String> allNodeClientIds) {
    }

    @Override
    public void onPreDir() {
        setPreviousRunningToSuccess(ExperimentState.SETUP_SOFTWARE, "Dir");
        setStatus(ExperimentState.SETUP_SOFTWARE, "Dir", ESpecStateItem.ESpecStateStatus.RUNNING);
    }

    @Override
    public void onPreDirNode(@Nonnull String nodeClientId) {
    }

    @Override
    public void onPostDir(@Nonnull DirSpec dir,
                          @Nonnull String fullPath,
                          @Nonnull String nodeClientId,
                          @Nonnull ESpecStepStatus status) {
        addLog().setText("Dir \""+fullPath+"\" Creation at "+nodeClientId).setStatus(status)
                .setESpecStepId("Dir")
                .add();
    }

    @Override
    public void onPostDirNode(@Nonnull String nodeClientId, boolean success) {
        addLog().setText("Finished Dir Creations at "+nodeClientId).setStatus(success)
                .setESpecStepId("Dir")
                .setIsMinor()
                .add();
    }

    @Override
    public void onPostDirAll(boolean success) {
        setStatus(ExperimentState.SETUP_SOFTWARE, "Dir", success ? ESpecStateItem.ESpecStateStatus.SUCCESS : ESpecStateItem.ESpecStateStatus.FAILURE);
        addLog().setText("Finished Dir Creations").setStatus(success)
                .setESpecStepId("Dir")
                .add();
    }

    @Override
    public void onPreUpload() {
        setPreviousRunningToSuccess(ExperimentState.SETUP_SOFTWARE, "Upload");
        setStatus(ExperimentState.SETUP_SOFTWARE, "Upload", ESpecStateItem.ESpecStateStatus.RUNNING);
    }

    @Override
    public void onPreUploadNode(@Nonnull String nodeClientId) {

    }

    @Override
    public void onPostUpload(@Nonnull UploadLikeSpec uploadSpec,
                             @Nonnull String fullPath,
                             @Nonnull String nodeClientId,
                             @Nonnull ESpecStepStatus status) {
        addLog().setText("Upload \""+fullPath+"\" at "+nodeClientId).setStatus(status)
                .setESpecStepId("Upload")
                .add();
    }

    @Override
    public void onPostUploadNode(@Nonnull String nodeClientId, boolean success) {
        addLog().setText("Finished Uploads at "+nodeClientId).setStatus(success)
                .setESpecStepId("Upload")
                .setIsMinor()
                .add();
    }

    @Override
    public void onPostUploadAll(boolean success) {
        setStatus(ExperimentState.SETUP_SOFTWARE, "Upload",
                success ? ESpecStateItem.ESpecStateStatus.SUCCESS : ESpecStateItem.ESpecStateStatus.FAILURE);
        addLog().setText("Finished Uploads").setStatus(success)
                .setESpecStepId("Upload")
                .add();
    }

    @Override
    public void onPreExecute() {
        setPreviousRunningToSuccess(ExperimentState.SETUP_SOFTWARE, "Execute");
        setStatus(ExperimentState.SETUP_SOFTWARE, "Execute", ESpecStateItem.ESpecStateStatus.RUNNING);
    }

    @Override
    public void onPostExecute(@Nonnull ExecuteSpec executeSpec,
                              @Nonnull String fullExecutablePath,
                              @Nullable String fullLogPath,
                              @Nonnull String nodeClientId,
                              int exitStatus,
                              @Nonnull ESpecStepStatus status) {
        addLog().setText("Execute \""+fullExecutablePath+"\" at "+nodeClientId).setStatus(status)
                .setESpecStepId("Execute")
                .add();
    }

    @Override
    public void onPostExecuteStepAll(@Nonnull ExecuteSpec executeSpec, boolean success) {

    }

    @Override
    public void onPostExecuteAll(boolean success) {
        setStatus(ExperimentState.SETUP_SOFTWARE, "Execute",
                success ? ESpecStateItem.ESpecStateStatus.SUCCESS : ESpecStateItem.ESpecStateStatus.FAILURE);
        addLog().setText("Finished Executes").setStatus(success)
                .setESpecStepId("Execute")
                .add();
    }

    @Override
    public void onPreAnsible() {
        setPreviousRunningToSuccess(ExperimentState.SETUP_SOFTWARE, "Ansible");
        setStatus(ExperimentState.SETUP_SOFTWARE, "Ansible", ESpecStateItem.ESpecStateStatus.RUNNING);
    }

    @Override
    public void onPostAnsiblePlaybook(@Nonnull AnsiblePlaybookSpec ansiblePlaybookSpec,
                                      @Nonnull String fullPlaybookPath,
                                      @Nullable String fullLogPath,
                                      @Nonnull String nodeClientId,
                                      @Nonnull ESpecStepStatus status) {
        addLog().setText("Ansible Playbook \""+fullPlaybookPath+"\" at "+nodeClientId).setStatus(status)
                .setESpecStepId("Ansible")
                .add();
    }

    @Override
    public void onPostAnsibleAll(boolean success) {
        setStatus(ExperimentState.SETUP_SOFTWARE, "Ansible",
                success ? ESpecStateItem.ESpecStateStatus.SUCCESS : ESpecStateItem.ESpecStateStatus.FAILURE);
        addLog().setText("Ansible Playbooks").setStatus(success)
                .setESpecStepId("Ansible")
                .add();
    }
}
