package be.iminds.ilabt.jfed.git;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.io.File;
import java.util.*;

public class GitFetcherCommandBuilder {
    private String gitUrl;
    private String repoSubDir;
    private String branch;
    private String username;
    private String password;
    private final Set<String> privateKeyPems = new HashSet<>();
    private File targetDir;
    private boolean allowCreateDir;
    private boolean overwriteFilesInDir;

    public GitFetcherCommandBuilder() {
    }

    public static Optional<GitFetcherCommandBuilder> fromString(@Nonnull String command) {
        boolean isSpecialGit = command.trim().startsWith("git ");
        boolean isSpecialGithub = command.trim().startsWith("github ");

        if (!isSpecialGit && !isSpecialGithub)
            return Optional.empty();

        String[] gitSourceSpecSplit = command.trim().split(" ");
        if (gitSourceSpecSplit.length > 4)
            throw new IllegalArgumentException(
                    "'"+command+"' has too many parts. Syntax:" +
                            " git[hub] https://<githubTLD>/<org>/<repo>.git [<subdir> [<branch>]]");
        if (gitSourceSpecSplit.length < 2)
            throw new IllegalArgumentException(
                    "'"+command+"' has too few parts. Syntax:" +
                            " git[hub] https://<githubTLD>/<org>/<repo>.git [<subdir> [<branch>]]");

        assert gitSourceSpecSplit[0].trim().startsWith("git");
        String gitUrl = gitSourceSpecSplit[1];
        String subDir = gitSourceSpecSplit.length > 2 ? gitSourceSpecSplit[2] : null;
        String branch = gitSourceSpecSplit.length > 3 ? gitSourceSpecSplit[3] : null;

        if (subDir != null && !subDir.endsWith("/"))
            subDir += "/";

        return Optional.of(new GitFetcherCommandBuilder()
                .setGitUrl(gitUrl)
                .setRepoSubDir(subDir)
                .setBranch(branch)
                .setAllowCreateDir(false)
                .setOverwriteFilesInDir(false));
    }
    
    public GitFetcherCommandBuilder(String gitUrl) {
        this.gitUrl = gitUrl;
    }

    public GitFetcherCommandBuilder setGitUrl(String gitUrl) {
        this.gitUrl = gitUrl;
        return this;
    }

    public GitFetcherCommandBuilder setRepoSubDir(String repoSubDir) {
        this.repoSubDir = repoSubDir;
        return this;
    }

    public GitFetcherCommandBuilder setBranch(String branch) {
        this.branch = branch;
        return this;
    }

    public GitFetcherCommandBuilder setUsername(String username) {
        this.username = username;
        return this;
    }

    public GitFetcherCommandBuilder setPassword(String password) {
        this.password = password;
        return this;
    }

    public GitFetcherCommandBuilder setPrivateKeyPems(Collection<String> privateKeyPems) {
        this.privateKeyPems.clear();
        this.privateKeyPems.addAll(privateKeyPems);
        return this;
    }

    public GitFetcherCommandBuilder addPrivateKeyPem(@Nonnull String privateKeyPem) {
        if (privateKeyPem == null)
            throw new IllegalArgumentException("addPrivateKeyPem(null)");
        this.privateKeyPems.add(privateKeyPem);
        return this;
    }

    public GitFetcherCommandBuilder setTargetDir(File targetDir) {
        this.targetDir = targetDir;
        return this;
    }

    public GitFetcherCommandBuilder setAllowCreateDir(boolean allowCreateDir) {
        this.allowCreateDir = allowCreateDir;
        return this;
    }

    public GitFetcherCommandBuilder setOverwriteFilesInDir(boolean overwriteFilesInDir) {
        this.overwriteFilesInDir = overwriteFilesInDir;
        return this;
    }

    public GitFetcherCommand build() {
        if (gitUrl == null)
            throw new IllegalStateException("gitUrl is a required argument of GitFetchCommand");
        File usedTargetDir = targetDir;
        if (usedTargetDir == null) {
            //fall back to a tmp dir
            usedTargetDir = GitFetcherCommand.createTempAnsibleDir();
        }
        return new GitFetcherCommand(gitUrl, repoSubDir, branch, username, password,
                new ArrayList<>(privateKeyPems),
                usedTargetDir, allowCreateDir, overwriteFilesInDir);
    }
}